try:
    from .BraceSequence import BraceSequence
    from .Sequence import Sequence
    from .ValueTerm import ValueTerm
except:
    from BraceSequence import BraceSequence
    from Sequence import Sequence
    from ValueTerm import ValueTerm

class ValueMap:
    def __init__(self) -> None:
        self.left = BraceSequence.init(Sequence.init([]))
        self.right = ValueTerm.init(Sequence.init([]))

    @classmethod
    def init(cls, left:BraceSequence, right:ValueTerm) -> 'ValueMap':
        if not isinstance(left, BraceSequence):
            raise TypeError()
        if not isinstance(right, ValueTerm):
            raise TypeError()
        new_item = ValueMap()
        new_item.left = left
        new_item.right = right
        return new_item
    
    # 这里右侧序列化的时候不使用中括号，但需要牢记右侧是 ValueTerm
    def serialize(self) -> str:
        return f"{self.left.serialize()} => {self.right.serialize()[1:-1]}"
    
    def json_obj(self) -> dict:
        return {
            "type": "ValueMap",
            "left": self.left.json_obj(),
            "right": self.right.json_obj()
        }
    
    @classmethod
    def deserialize(cls, s:str) -> 'ValueMap':
        if s.find("=>") == -1:
            raise ValueError()
        left_str, right_str = s.split("=>", maxsplit=1)
        new_item = ValueMap()
        new_item.left = BraceSequence.deserialize(left_str)
        new_item.right = ValueTerm.deserialize(f"[{right_str}]")
        return new_item
    
    @classmethod
    def from_json_obj(cls, json_obj:dict) -> 'ValueMap':
        if json_obj.get("type") != "ValueMap":
            raise TypeError()
        if json_obj.get("left") is None:
            raise ValueError()
        if json_obj.get("right") is None:
            raise ValueError()
        new_item = ValueMap()
        new_item.left = BraceSequence.from_json_obj(json_obj["left"])
        new_item.right = ValueTerm.from_json_obj(json_obj["right"])
        return new_item
    
if __name__ == "__main__":
    
    test_list = [
        "() => a b c",
        "(a (b c d) e) => (a (b) c)",
        "(a b) => a"
    ]
    for value in test_list:
        print(ValueMap.deserialize(value).serialize())
        print(ValueMap.from_json_obj(ValueMap.deserialize(value).json_obj()).serialize())



from typing import Optional

try:
    from .SimpleTerm import SimpleTerm
    from .BraceSequence import BraceSequence
    from .Sequence import Sequence
except:
    from SimpleTerm import SimpleTerm
    from BraceSequence import BraceSequence
    from Sequence import Sequence

class ValueTerm:
    def __init__(self) -> None:
        self.value = Sequence.init([])

    @classmethod
    def init(cls, value:BraceSequence|Sequence) -> 'ValueTerm':
        if type(value) not in [BraceSequence, Sequence]:
            raise TypeError()
        new_item = ValueTerm()
        new_item.value = value
        return new_item
    
    def serialize(self) -> str:
        return f"[{self.value.serialize()}]"
    
    def json_obj(self) -> dict:
        return {
            "type": "ValueTerm",
            "value": self.value.json_obj()
        }
    
    @classmethod
    def deserialize(cls, s:str) -> 'ValueTerm':
        s = s.strip()
        if not s.startswith("["):
            raise ValueError()
        if not s.endswith("]"):
            raise ValueError()
        s = s[1:-1].strip()

        # 为了正确序列化匹配的括号
        # 我们这里先假设对象是一个 Sequence
        value = Sequence.deserialize(s)

        # 如果发现这个 Sequence 中唯一的元素就是 BraceSequence
        # 再把这个元素拿出来
        if len(value.objects) == 1 and isinstance(value.objects[0], BraceSequence):
            value = value.objects[0]
        
        new_item = ValueTerm()
        new_item.value = value
        return new_item
    
    @classmethod
    def from_json_obj(cls, json_obj:dict) -> 'ValueTerm':
        if json_obj.get("type") != "ValueTerm":
            raise TypeError()
        if not isinstance(json_obj.get("value"), dict):
            raise TypeError()
        value_json_obj = json_obj["value"]
        if value_json_obj.get("type") not in ["BraceSequence", "Sequence"]:
            raise TypeError()
        new_item = ValueTerm()
        if value_json_obj["type"] == "Sequence":
            new_item.value = Sequence.from_json_obj(value_json_obj)
        else:
            new_item.value = BraceSequence.from_json_obj(value_json_obj)
        return new_item
    
    # 获取一个当前序列中的任意变量名
    # 找不到则返回 None
    def get_one_var(self) -> Optional[str]:
        return self.value.get_one_var()
    
    # 一种简单的数据表达方式
    # 在 python 传参接口中使用
    def simple_express(self) -> list:
        if isinstance(self.value, BraceSequence):
            return [self.value.simple_express()]
        else:
            if not isinstance(self.value, Sequence):
                raise TypeError()
            return self.value.simple_express()

if __name__ == "__main__":
    test_list = [
        "[]",            # 没有元素 Sequence
        "[a]",           # 单个元素 Sequence
        "[a (b c) d]",   # 三个元素 Sequence
        "[(a (b c) d)]", # 一个 BraceSequence
        "[(a b) (c d)]", # 多个 BraceSequence
    ]
    for value in test_list:
        value_term = ValueTerm.deserialize(value)
        print(value_term.serialize())
        print(ValueTerm.from_json_obj(value_term.json_obj()).serialize())
        print(value_term.simple_express())
        print()



import string

try:
    from .BraceSequence import BraceSequence
    from .SimpleTerm import SimpleTerm
    from .Sequence import Sequence
except:
    from BraceSequence import BraceSequence
    from SimpleTerm import SimpleTerm
    from Sequence import Sequence

# b_tmp 是模式串
# b_txt 是文本串，文本串中不可以有星号，模式串可以有
def match_brace_sequence(
        b_tmp:BraceSequence, 
        b_txt:BraceSequence, dic:dict[str, Sequence]) -> bool:
    
    if not isinstance(b_tmp, BraceSequence): # 检查类型
        raise TypeError()
    if not isinstance(b_txt, BraceSequence):
        raise TypeError()

    # 两个正在参与对比的序列
    tmp_seq = b_tmp.inner_sequence.objects
    txt_seq = b_txt.inner_sequence.objects

    # 长度为零的序列只能和长度为 0 的序列匹配
    if len(tmp_seq) == 0:
        return len(txt_seq) == 0

    # 对当前两个括号表达式试图进行前缀匹配
    index = 0
    while index < len(tmp_seq):

        # 模板串中当前元素是 SimpleTerm 的情况
        if isinstance(tmp_seq[index], SimpleTerm):
            if SimpleTerm.is_const_val(tmp_seq[index].serialize()): # 强制匹配
                if index >= len(txt_seq): # 强制匹配时长度不足
                    return False
                elif tmp_seq[index].serialize() != txt_seq[index].serialize(): # 匹配失败
                    return False
                else:
                    index += 1
                    pass # 成功匹配一个项目
            
            elif tmp_seq[index].has_star: # 匹配后面全部内容
                if SimpleTerm.is_const_val(tmp_seq[index].serialize()[1:]):
                    raise ValueError(f"Star match \"{tmp_seq[index].serialize()}\" should use variable name not constant name \"{tmp_seq[index].serialize()[1:]}\".")

                # 通配符只能出现在右括号前
                if index != len(tmp_seq) - 1:
                    raise ValueError()
                
                # 匹配所有内容
                arr = []
                for i in range(index, len(txt_seq)):
                    arr.append(txt_seq[i])

                # 变量名重复
                if dic.get(tmp_seq[index].serialize()) is not None:
                    raise ValueError()

                dic[tmp_seq[index].serialize()] = Sequence.init(arr)
                return True
            
            # 匹配单个符号的变量，恰好可以匹配一个常量符号或者一个括号对象
            # 恰好匹配一个元素
            else: 

                # 如果文本串中没有可以匹配的元素
                # 则不可能构成匹配（匹配失败）
                if index >= len(txt_seq):
                    return False
                
                # 变量名重复
                if dic.get(tmp_seq[index].serialize()) is not None:
                    raise ValueError()
                
                # 这里即使是一个元素，也应该是序列
                dic[tmp_seq[index].serialize()] = Sequence.init([
                    txt_seq[index]
                ])

                index += 1
                pass # 成功匹配了一个条目
            
        # 模板串中当前元素是 BraceSequence 的情况
        else:

            # 此处必须是括号表达式
            if not isinstance(tmp_seq[index], BraceSequence):
                raise ValueError()
            if (index >= len(txt_seq)) or not isinstance(txt_seq[index], BraceSequence):
                return False
            
            flag = match_brace_sequence(
                tmp_seq[index], txt_seq[index], dic)
            if not flag:
                return False # 内层匹配失败了
            
            # 程序执行到这里说明内层匹配成功了
            index += 1
            pass
    
    if index != len(txt_seq): # 说明 txt_seq 有值域没有消耗完
        return False
    return True

if __name__ == "__main__":
    for tmp_str, txt_str in [
        ("(HEAD (a *b))", "(HEAD (1 2 3 4 5))"),
        ("()", "(1 2 3 4 5)")
    ]:
        dic = dict()
        tmp = BraceSequence.deserialize(tmp_str)
        txt = BraceSequence.deserialize(txt_str)

        flag = match_brace_sequence(tmp, txt, dic)
        print(flag)

        for var_name in dic:
            print(f"{var_name}: {dic[var_name].serialize()}")
        print()



import functools
import pip_pkg_info

@functools.cache
def get_version_in_toml() -> str:
    import os
    DIRNOW = os.path.dirname(os.path.abspath(__file__))
    TOML_PATH = os.path.join(DIRNOW, "..", "pyproject.toml")

    if not os.path.isfile(TOML_PATH):
        return "unknown"

    for line in open(TOML_PATH):
        line = line.strip()
        if line.find("version") != -1: # 找到 version
            if line[-1] == "\"":
                return line[1:-1].split("\"")[-1]
    
    return "unknown"

# 获得当前软件的版本
@functools.cache
def get_version(pkg_name_now:str="xulang") -> str:
    dic = pip_pkg_info.pip_pkg_info()

    if dic.get(pkg_name_now) is not None:
        return dic[pkg_name_now]["version"]
    
    else:
        return get_version_in_toml()

if __name__ == "__main__":
    print(get_version())



from typing import Optional

try:
    from .ValueMap import ValueMap
    from .ValueTerm import ValueTerm
    from .MatchBraceSequence import match_brace_sequence
    from .FillValueTerm import fill_value_term
    from .Sequence import Sequence
    from .BraceSequence import BraceSequence
    from .SimpleTerm import SimpleTerm
except:
    from ValueMap import ValueMap
    from ValueTerm import ValueTerm
    from MatchBraceSequence import match_brace_sequence
    from FillValueTerm import fill_value_term
    from Sequence import Sequence
    from BraceSequence import BraceSequence
    from SimpleTerm import SimpleTerm

class ValueMapWrap:
    def __init__(self, index:int, value_map:ValueMap, filepath:str, line_id:int) -> None:
        self.value_map = value_map
        self.filepath = filepath
        self.line_id = line_id
        self.index = index # index 决定重新排序后的顺序

class RuleSet:
    def __init__(self) -> None:

        # self.value_map_dict 是用于快速检索可行命令的关键
        # 该 dict 构建了一个映射：从第一个元素映射到对应的命令序列
        # 如果某个 value_map.left 的第一个元素是常量，则应该放入这个常量对应的序列
        # 否则应该放入 "" 对应的序列
        # "" 对应的序列表示所有 value_map.left 中没有元素或者第一个元素不是常量的规则
        self.value_map_dict:dict[str, list[ValueMapWrap]] = dict()

        # 记录最后一次匹配所使用的命令信息
        self.latest_used_rule:Optional[ValueMapWrap] = None

        # 记录当前规则集合中总共有多少条规则，用于给规则排序
        self.rule_count_now = 0
    
    def safe_append(self, dic: dict[str, list[ValueMapWrap]], key:str, val:ValueMapWrap):
        if dic.get(key) is None:
            dic[key] = []
        dic[key].append(val)

    def add_rule(self, value_map: ValueMap, filepath:str, line_id:int): # 新增规则
        if not isinstance(value_map, ValueMap):
            raise TypeError()
        if not isinstance(value_map.left, BraceSequence):
            raise TypeError()

        # 获得新的规则编号
        self.rule_count_now += 1
        
        bad = False

        # 先考虑没有任何元素的情况
        item_cnt = len(value_map.left.inner_sequence.objects)
        if (not bad) and item_cnt == 0:
            bad = True
        
        # 试图获取第一个元素（如果没有元素则获取到空）
        if not bad:
            first_item = value_map.left.inner_sequence.objects[0]
        else:
            first_item = None
        
        # 有元素但是第一个元素不是 SimpleTerm
        if (not bad) and not isinstance(first_item, SimpleTerm):
            if not isinstance(first_item, BraceSequence):
                raise TypeError()
            bad = True
        
        # 第一个元素是 SimpleTerm 但不是常量
        assert isinstance(first_item, SimpleTerm)
        if (not bad) and not SimpleTerm.is_const_val(first_item.serialize()):
            bad = True
        
        # 第一个元素不符合快速检索规则
        if bad:
            self.safe_append(
                self.value_map_dict, "", # 放入空字符串对应的序列
                ValueMapWrap(self.rule_count_now, value_map, filepath, line_id))
        
        # 第一个元素符合快速检索规则
        else:
            if not isinstance(first_item, SimpleTerm):
                raise TypeError()
            if not SimpleTerm.is_const_val(first_item.serialize()):
                raise ValueError()
            self.safe_append(
                self.value_map_dict, first_item.serialize(), # 放入空字符串对应的序列
                ValueMapWrap(self.rule_count_now, value_map, filepath, line_id))

    # 获取所有可以使用的命令
    # 在获取命令时根据第一个元素进行初筛
    # 这种初筛可以显著提升算法效率
    def get_all_value_map_wrap(self, first_item:Optional[str]) -> list[ValueMapWrap]:
        arr = []

        # first_item 为 None 表示不对第一个符号进行快速筛查
        if first_item is None:
            for item in self.value_map_dict:
                arr += self.value_map_dict[item]

        # 需要对第一个符号进行快速筛查
        # 注意记得考虑 "" 对应的所有规则也需要参与比较
        else:
            for key_term in set([first_item, ""]): # 需要使用 set 去重，因为 first_item 可能是 ""
                if self.value_map_dict.get(key_term) is not None:
                    arr += self.value_map_dict[key_term]

        # 按照插入顺序排序（保证优先级正确）
        return sorted(arr, key=lambda x:x.index)

    # 显示所有的规则
    def show_rules(self, prefix:Optional[str]) -> str:
        int_width = len(str(len(self.get_all_value_map_wrap(None)))) + 1 # 编号需要的宽度
        arr = [
            ((
                ("%" + str(int_width) + "d ")
                    % value_map_wrap.index) 
                + value_map_wrap.value_map.serialize() + "\n"
            )
            for value_map_wrap in self.get_all_value_map_wrap(prefix)
        ]
        return "".join(arr)
    
    # 删除某个前缀对应的所有规则
    def erase_rule(self, prefix:str):
        if self.value_map_dict.get(prefix) is not None:
            del self.value_map_dict[prefix]
    
    # 根据 value_term 中内容的第一个元素进行快速分类
    def select_first_term_key(self, value_term:ValueTerm):
        if not isinstance(value_term.value, BraceSequence):
            raise AssertionError()
        
        # 如果其中没有元素
        if len(value_term.value.inner_sequence.objects) == 0:
            return ""
        
        # 如果其中有至少一个元素
        # 如果这个元素是复杂元素，也不可能和任意常量进行匹配
        first_item = value_term.value.inner_sequence.objects[0]
        if not isinstance(first_item, SimpleTerm):
            return ""
        
        # 理论上 value_term 中不可以存在变量，因此只要有 SimpleTerm 就一定是常量
        if not SimpleTerm.is_const_val(first_item.serialize()):
            raise AssertionError()
        return first_item.serialize()

    # 执行一次所有简单规则
    def execute_simple_rules(self, value_term:ValueTerm) -> tuple[bool, ValueTerm]:
        if not isinstance(value_term.value, BraceSequence):
            return False, value_term
        
        # 计算初筛分类
        first_term_key = self.select_first_term_key(value_term)

        # 根据初筛分类快速排除对象
        for value_map_wrap in self.get_all_value_map_wrap(first_term_key):
            value_map = value_map_wrap.value_map

            # 检查是否还能替换
            if isinstance(value_term.value, Sequence):
                break

            assert isinstance(value_map.left, BraceSequence)
            if value_map.left.has_sub_brace(): # 跳过复杂规则
                continue

            dic = dict()
            flag = match_brace_sequence(
                value_map.left, value_term.value, dic)
            
            if flag:
                value_term = fill_value_term(
                    value_map.right, dic
                )
                self.latest_used_rule = value_map_wrap
                return True, value_term

        # 返回最终结果
        return False, value_term

    def try_match_sons(self, value_term:ValueTerm) -> tuple[bool, ValueTerm]:
        if isinstance(value_term.value, BraceSequence):
            sub_obj_list = value_term.value.inner_sequence.objects
            value_type = BraceSequence
        else:
            sub_obj_list = value_term.value.objects
            value_type = Sequence

        suc = False
        new_sub_obj_list = []
        for i in range(len(sub_obj_list)):
            if type(sub_obj_list[i]) not in [BraceSequence, Sequence] or suc:
                new_sub_obj_list.append(sub_obj_list[i])
                continue
            flag, new_obj = self.calc_once(ValueTerm.init(sub_obj_list[i]))

            if isinstance(new_obj.value, BraceSequence):
                new_sub_obj_list.append(new_obj.value)
            else:
                assert isinstance(new_obj.value, Sequence)
                new_sub_obj_list += new_obj.value.objects # 追加模式

            if flag:
                suc = True # 至少进行了一次替换

        if value_type is BraceSequence:
            return suc, ValueTerm.init(BraceSequence.init(
                Sequence.init(new_sub_obj_list)
            ))
        else:
            return suc, ValueTerm.init(
                Sequence.init(new_sub_obj_list)
            )

    # 计算一个表达式的最终结果
    # verbose 模式可以显示计算过程
    def calc(self, value_term:ValueTerm, verbose:bool, step_mode:bool) -> ValueTerm:
        if verbose:
            print("FROM:", value_term.serialize()[1:-1])
        while True:
            flag, value_term = self.calc_once(value_term)
            if flag == False:
                break
            if self.latest_used_rule is not None:
                if verbose:
                    print("\nRULE:", self.latest_used_rule.value_map.serialize())
            if verbose:
                print("OUTP:", value_term.serialize()[1:-1])
                if step_mode:
                    input("(Continue?) ") # press enter to continue
        return value_term

    # 使用当前规则集合进行必要的计算
    def calc_once(self, value_term:ValueTerm) -> tuple[bool, ValueTerm]:

        if type(value_term.value) not in [Sequence, BraceSequence]:
            raise TypeError()

        # 先考虑外层是括号的情况
        if isinstance(value_term.value, BraceSequence):
            
            # 先枚举无嵌套规则并执行
            # 此时当前层已经不再能执行
            flag, value_term = self.execute_simple_rules(value_term)
            if flag:
                return True, value_term

            # 试图递归对所有成员执行算法
            flag, value_term = self.try_match_sons(value_term)
            if flag:
                return True, value_term
            
            
            # 试图对当前层进行复杂规则替换
            # 此时仍然需要根据第一个对象进行快速初筛
            first_item_key = self.select_first_term_key(value_term)
            for value_map_wrap in self.get_all_value_map_wrap(first_item_key):
                value_map = value_map_wrap.value_map
                
                assert isinstance(value_map, ValueMap)
                assert isinstance(value_term.value, BraceSequence)

                if value_map.left.has_sub_brace(): # 复杂规则
                    dic = dict()
                    flag = match_brace_sequence(
                        value_map.left, value_term.value, dic)
                    
                    if flag:
                        value_term = fill_value_term(value_map.right, dic)
                        self.latest_used_rule = value_map_wrap
                        return True, value_term
            
            return False, value_term
        
        else:

            # 直接对子元素进行递归即可
            if not isinstance(value_term.value, Sequence):
                raise AssertionError()
            
            # 试图递归对所有成员执行算法
            flag, value_term = self.try_match_sons(value_term)
            if flag:
                return True, value_term

            return False, value_term

if __name__ == "__main__":
    rule_set = RuleSet()
    rule_set.add_rule(ValueMap.deserialize("""
        (IF TRUE a b) => a"""), "<TEST>", 1)
    rule_set.add_rule(ValueMap.deserialize("""
        (IF FALSE a b) => b"""), "<TEST>", 2)
    rule_set.add_rule(ValueMap.deserialize("""
        (HEAD ()) => """), "<TEST>", 3)
    rule_set.add_rule(ValueMap.deserialize("""
        (HEAD (a *b)) => a"""), "<TEST>", 4)
    rule_set.add_rule(ValueMap.deserialize("""
        (REV ()) => ()"""), "<TEST>", 5)    
    rule_set.add_rule(ValueMap.deserialize("""
        (REV (a)) => (a)"""), "<TEST>", 6)
    rule_set.add_rule(ValueMap.deserialize("""
        (REV (a *b)) => (MERGE (REV (*b)) (a))"""), "<TEST>", 7)
    rule_set.add_rule(ValueMap.deserialize("""
        (MERGE (*a) (*b)) => (*a *b)"""), "<TEST>", 8)
    rule_set.add_rule(ValueMap.deserialize("""
        (TAIL (*a)) => (HEAD (REV (*a)))"""), "<TEST>", 9)
    
    for value_term_str in [
        "(IF TRUE 1 (TAIL (A B C D E)))",
        "(IF FALSE X (TAIL (A B C D E)))"
    ]:
        print("=" * 40)
        value_term = ValueTerm.deserialize(f"[{value_term_str}]")
        rule_set.calc(value_term, True, False)



from typing import Optional

try:
    from .Sequence import Sequence, BRACE_SEQUENCE_CLASS_META_OBJECT
except:
    from Sequence import Sequence, BRACE_SEQUENCE_CLASS_META_OBJECT

class BraceSequence:
    def __int__(self):
        self.inner_sequence = Sequence.init([])
    
    @classmethod
    def init(cls, inner_sequence: Sequence) -> 'BraceSequence':
        if not isinstance(inner_sequence, Sequence):
            raise TypeError()
        new_obj = BraceSequence()
        new_obj.inner_sequence = inner_sequence
        return new_obj
    
    def serialize(self) -> str:
        return f"({self.inner_sequence.serialize()})"
    
    
    def json_obj(self) -> dict:
        return {
            "type": "BraceSequence",
            "inner_sequence": self.inner_sequence.json_obj()
        }
    
    @classmethod
    def deserialize(cls, s:str) -> 'BraceSequence':
        s = s.strip()
        if not s.startswith("("):
            raise ValueError()
        if not s.endswith(")"):
            raise ValueError("Unterminated parenthetical expression.")
        new_obj = BraceSequence()
        new_obj.inner_sequence = Sequence.deserialize(s[1:-1])
        return new_obj
    
    @classmethod
    def from_json_obj(cls, json_obj:dict) -> 'BraceSequence':
        if json_obj.get("type") != "BraceSequence":
            raise TypeError()
        if json_obj.get("inner_sequence") is None:
            raise ValueError()
        new_obj = BraceSequence()
        new_obj.inner_sequence = Sequence.from_json_obj(
            json_obj["inner_sequence"]
        )
        return new_obj
    
    def has_sub_brace(self) -> bool:
        return self.inner_sequence.has_sub_brace()
    
    # 获取一个当前序列中的任意变量名
    # 找不到则返回 None
    def get_one_var(self) -> Optional[str]:
        return self.inner_sequence.get_one_var()
    
    # 一种简单的使用 tuple 和 list of str 表示数据的方式
    # 在 python 传参接口中使用
    def simple_express(self) -> tuple[str]:
        return tuple(self.inner_sequence.simple_express()) # type:ignore

# 传参方式传递类型对象
BRACE_SEQUENCE_CLASS_META_OBJECT[0] = BraceSequence # type: ignore

if __name__ == "__main__":
    item = BraceSequence.deserialize("(a (b (*c) *f d) e)")
    print(item.serialize())
    print(BraceSequence.from_json_obj(item.json_obj()).serialize())

    print(item.simple_express())



try:
    from .Sequence import Sequence
    from .BraceSequence import BraceSequence
    from .ValueTerm import ValueTerm
    from .SimpleTerm import SimpleTerm
except:
    from Sequence import Sequence
    from BraceSequence import BraceSequence
    from ValueTerm import ValueTerm
    from SimpleTerm import SimpleTerm

def fill_value_term(right:ValueTerm, dic:dict[str, Sequence]) -> ValueTerm:
    if not isinstance(right, ValueTerm):
        raise TypeError()
    if isinstance(right.value, Sequence):
        new_item = ValueTerm()
        new_item.value = fill_sequence(right.value, dic)
        return new_item
    else:
        if not isinstance(right.value, BraceSequence):
            raise TypeError()
        new_item = ValueTerm()
        new_item.value = fill_brace_sequence(right.value, dic) # type:ignore
        return new_item

def fill_sequence(sequence:Sequence, dic:dict[str, Sequence]) -> Sequence:
    object_list = []
    for term in sequence.objects:
        if isinstance(term, SimpleTerm):
            
            # 遇到变量
            if not SimpleTerm.is_const_val(term.serialize()):
                if dic.get(term.serialize()) is None:
                    raise NameError()
                
                # 追加序列
                object_list += (dic[term.serialize()].objects)

            # 遇到常量
            else:
                object_list.append(term)
        else:
            if not isinstance(term, BraceSequence):
                raise TypeError()
            
            object_list.append(
                fill_brace_sequence(term, dic))
            
    new_item = Sequence.init(object_list)
    return new_item

def fill_brace_sequence(brace_sequence:BraceSequence, dic:dict[str, Sequence]) -> BraceSequence:
    return BraceSequence.init(
        fill_sequence(brace_sequence.inner_sequence, dic))

if __name__ == "__main__":
    print(
        fill_value_term(ValueTerm.init(
            Sequence.deserialize("(a (b A) c)")),
            {
                "a": Sequence.deserialize("A B"),
                "b": Sequence.deserialize(""),
                "c": Sequence.deserialize("(E F) G")
            }).serialize()
    )



from typing import Optional

try:
    from .SimpleTerm import SimpleTerm
except:
    from SimpleTerm import SimpleTerm

def flush(buffer:str, arr:list) -> str:
    if buffer != "":
        arr.append(buffer)
    return ""

def smart_split(s:str) -> list[str]:
    arr = []
    buffer = ""
    for i in range(len(s)):
        if s[i].isspace():
            buffer = flush(buffer, arr)
        elif s[i] in ["(", ")"]:
            buffer = flush(buffer, arr)
            arr.append(s[i])
        else:
            buffer += s[i]
    buffer = flush(buffer, arr)
    return arr

def find_match(str_list:list[str], index:int):
    cnt = 0
    pos = -1
    for j in range(index, len(str_list)):
        if str_list[j] == "(":
            cnt += 1
        elif str_list[j] == ")":
            cnt -= 1
        if cnt == 0 and j > index:
            pos = j
            break
    if pos == -1:
        raise ValueError("Unpaired parenthesis.") # 没有找到匹配的括号
    return pos

BRACE_SEQUENCE_CLASS_META_OBJECT = [None]

class Sequence:
    def __init__(self) -> None:
        self.objects = []
    
    @classmethod
    def init(cls, objects:list) -> 'Sequence':
        for term in objects:
            if type(term).__name__ not in ["SimpleTerm", "BraceSequence"]:
                raise TypeError()
        new_item = Sequence()
        new_item.objects = objects
        return new_item
    
    def serialize(self) -> str:
        return " ".join([
            term.serialize() for term in self.objects
        ])
    
    def json_obj(self) -> dict:
        return {
            "type": "Sequence",
            "objects": [
                term.json_obj() for term in self.objects
            ]
        }
    
    @classmethod
    def deserialize(cls, s:str) -> 'Sequence':
        if BRACE_SEQUENCE_CLASS_META_OBJECT[0] is None:
            raise AssertionError()
        str_list = smart_split(s)
        object_list = []
        index = 0
        while index < len(str_list):
            if str_list[index] == ")":
                raise ValueError("Unexpected closing parenthesis.")
            elif str_list[index] == "(":
                pos = find_match(str_list, index)
                object_list.append(
                    BRACE_SEQUENCE_CLASS_META_OBJECT[0].deserialize(
                        " ".join(str_list[index: pos+1])
                    )
                )
                index = pos + 1
            else:
                object_list.append(
                    SimpleTerm.deserialize(str_list[index])
                )
                index += 1
        return Sequence.init(object_list)

    @classmethod
    def from_json_obj(cls, json_obj:dict) -> 'Sequence':
        if BRACE_SEQUENCE_CLASS_META_OBJECT[0] is None:
            raise AssertionError()
        if json_obj.get("type") != "Sequence":
            raise TypeError()
        if not isinstance(json_obj.get("objects"), list):
            raise TypeError()
        object_list = []
        for term in json_obj["objects"]:
            if term.get("type") not in ["SimpleTerm", "BraceSequence"]:
                raise TypeError()
            if term["type"] == "SimpleTerm":
                object_list.append(
                    SimpleTerm.from_json_obj(term)
                )
            else:
                object_list.append(
                    BRACE_SEQUENCE_CLASS_META_OBJECT[0].from_json_obj(term)
                )
        return Sequence.init(object_list)
    
    # 检查当前序列是否有一个成员是 BraceSequence
    # 没有 BraceSequence 成员的 Sequence 是 “简单的”
    # 否则它就是 “复杂的”
    # 简单规则总是优先从顶层开始匹配，先匹配（用于保证短路性）
    # 复杂规则总是从底层开始匹配
    def has_sub_brace(self) -> bool:
        if BRACE_SEQUENCE_CLASS_META_OBJECT[0] is None:
            raise AssertionError()
        for item in self.objects:
            if isinstance(item, BRACE_SEQUENCE_CLASS_META_OBJECT[0]):
                return True
        return False

    # 获取一个当前序列中的任意变量名
    # 找不到则返回 None
    def get_one_var(self) -> Optional[str]:
        for term in self.objects:
            one_var = term.get_one_var()
            if one_var is not None:
                return one_var
        return None
    
    # 一种简单的使用 tuple 和 list of str 表示数据的方式
    # 在 python 传参接口中使用
    def simple_express(self) -> list[str]:
        return [
            term.simple_express()
            for term in self.objects
        ]

if __name__ == "__main__":
    if BRACE_SEQUENCE_CLASS_META_OBJECT[0] is None:
        from BraceSequence import BraceSequence
        BRACE_SEQUENCE_CLASS_META_OBJECT[0] = BraceSequence # type: ignore
    
    for item_str in [
        "a (x y) (z)",
        "(x y) (p q)"
    ]:
        sequence_item = Sequence.deserialize(item_str)
        print(sequence_item.serialize())
        print(Sequence.from_json_obj(sequence_item.json_obj()).serialize())



from .BraceSequence import BraceSequence
from .FileRunner import FileRunner
from .RuleSet import RuleSet
from .Sequence import Sequence
from .SimpleTerm import SimpleTerm
from .ValueMap import ValueMap
from .ValueTerm import ValueTerm
from .GetVersion import get_version

__all__ = [
    "BraceSequence",
    "FileRunner",
    "RuleSet",
    "Sequence",
    "SimpleTerm",
    "ValueMap",
    "ValueTerm",
    "get_version"
]



import sys
sys.setrecursionlimit(5000) # 设置最大递归深度

import os
import json
from . import FileRunner, get_version

# 从命令行中拆分出所有 -I 命令处理
def split_include_path(argv_list:list[str]) -> tuple[list[str], list[str]]:
    CMD_PREFIX = "-I"
    index = 0
    new_argv_list = []     # 删除 -I 命令后的其他命令
    include_path_list = [] # 拆分出所有 include_path
    while index < len(argv_list):

        # 检测到了 -I 的命令
        if argv_list[index].startswith(CMD_PREFIX):

            # 单纯的 -I 命令：下一个对象将会是路径描述符
            if len(argv_list[index]) == len(CMD_PREFIX):
                if (index + 1) >= len(argv_list):
                    raise ValueError("\"-I\" should be followed by a path.")
                include_path_list.append(argv_list[index + 1].strip())
                index += 2

            # 自身带有路径的 -I 命令
            else:
                if len(argv_list[index]) <= len(CMD_PREFIX): # 一定更长
                    raise AssertionError()
                other_part = argv_list[index][len(CMD_PREFIX):] # 删除前缀
                include_path_list.append(other_part.strip())
                index += 1
        
        # 检测到了无关命令
        else: 
            new_argv_list.append(argv_list[index])
            index += 1
    
    # 返回拆分后的结果
    # 注意 "-I." 以及 "-I.." 之类的写法需要改写成当前工作目录相对路径
    return new_argv_list, [
        os.path.abspath(os.getcwd()) if include_path == "." else (
        os.path.dirname(os.path.abspath(os.getcwd())) if include_path == ".." else include_path)
        for include_path in include_path_list
    ]

# 检查某个字符串是否存在
# 并返回删除该字符串后的结果
def check_and_erase(argv_list:list[str], aim_str:str) -> tuple[list[str], bool]:
    ans = aim_str in argv_list
    new_argv_list = [
        item
        for item in argv_list
        if item != aim_str
    ]
    return new_argv_list, ans

def main(argv_list:list[str]) -> int:

    # 检查 -I 命令
    # 这个命令用于指定 include 路径
    try:
        argv_list, inlcude_path_list = split_include_path(argv_list)
    except ValueError as e:
        print("<CMD:1>", e)
        return 1

    # 检查 --verbose：是否需要逐条输出匹配过程
    argv_list, verbose = check_and_erase(argv_list, "--verbose")
    argv_list, step_mode = check_and_erase(argv_list, "--step")

    # --step 的功能，是在开启 --verbose 的前提下
    # 每次替换后，需要按一下 Enter 才能继续执行，方便观察代码运行是否正确
    if step_mode == True and verbose == False:
        print("<CMD>:1", "\"--step\" can only work with \"--verbose\".")
        return 1

    # 初始化文件运行器
    try:
        file_runner = FileRunner(inlcude_path_list)
        file_runner.verbose = verbose
        file_runner.step_mode = step_mode

    # 在初始化过程中如果出现错误
    # 一般来说是 -I 指定的 include_path 不存在
    except Exception as e:
        print("<CMD>:1", e)
        return 1

    # 参数中没有指定需要运行的文件
    # 启动交互式命令行
    if argv_list == []:
        print(f"xulang interactive command line v{get_version()}")
        file_runner.interactive_ui()

    # 检查命令行中指定的文件个数是否超过一个
    # 如果超过一个，则报错，并显示帮助信息
    elif len(argv_list) > 1:
        print("Usage:")
        print("    python3 -m xulang")
        print("    python3 -m xulang <source_file>")
        print("    python3 -m xulang -I <include_path>")
        print("    python3 -m xulang --verbose")
        print("    python3 -m xulang --verbose --step")
        return 1

    # 命令中恰好指定了一个需要执行的文件时
    # 执行这个文件
    else:
        # 检查需要执行的文件是否存在
        filepath = argv_list[0]
        if not os.path.isfile(filepath):
            print(f"<CMD>:1 File \"{filepath}\" not found!")
            return 1
        
        # file_runner.run_file 需要使用绝对路径调用程序
        # 需要将非绝对路径转换为绝对路径
        if not os.path.isabs(filepath): 
            filepath = os.path.abspath(filepath)
        file_runner.run_file(filepath)

    # 返回 0 表示程序自正确结束
    return 0

# 运行主程序
sys.exit(main(json.loads(json.dumps(sys.argv[1:]))))



import os
import re
import traceback
from typing import Optional

try:
    from .RuleSet import RuleSet
    from .ValueMap import ValueMap
    from .ValueTerm import ValueTerm
    from .Sequence import Sequence
    from .BraceSequence import BraceSequence
    from .MatchBraceSequence import match_brace_sequence
    from .VarSet import VarSet
    from .SimpleTerm import SimpleTerm
except:
    from RuleSet import RuleSet
    from ValueMap import ValueMap
    from ValueTerm import ValueTerm
    from Sequence import Sequence
    from BraceSequence import BraceSequence
    from MatchBraceSequence import match_brace_sequence
    from VarSet import VarSet
    from .SimpleTerm import SimpleTerm

# 当前目录
DIRNOW = os.path.dirname(
    os.path.abspath(__file__))

# 标准库的 include 文件
INCLUDE_DIR = os.path.join(DIRNOW, 
    "include")

class CommandWrap:
    def __init__(self, filepath:str, line_id:int, command:str) -> None:
        self.filepath = filepath
        self.line_id = line_id
        self.command = command

class FileRunner:
    def __init__(self, include_path_list:list[str]) -> None:
        if not isinstance(include_path_list, list): # 必须是 list
            raise TypeError()
        for include_path in include_path_list:
            if not isinstance(include_path, str): # 必须是字符串
                raise TypeError()
            if not os.path.isdir(include_path):
                raise FileNotFoundError(f"Include path \"{include_path}\" not found.")

        # 记录已经被引入的所有路径
        # 保证同一个文件第二次 include 时不再执行
        self.exists_path:set[str] = set()

        # 记录所有还没有执行的指令
        self.cmd_list:list[CommandWrap] = []

        # 记录目前所有的匹配规则
        self.rule_set = RuleSet()

        # 当前指令执行是否需要输出（手动设置为 True 可以执行命令时显示执行路径）
        self.verbose = False

        # 设置单步暂停模式
        self.step_mode = False

        # 报错是否需要打开源代码位置信息
        self.extra_error_info = False

        # 记录所有可以 include 的路径
        #   这里的 "." 是一个特殊路径，表示被运行的脚本所在的目录
        #   而不是当前工作路径
        # self.include_path 中除了 "." 之外都是绝对路径
        # 一定要把标准库 INCLUDE_DIR 放在最后
        #   因为 include_path 越靠前优先级越高（优先级最高的永远是相对路径 include）
        self.include_path = ["."] + [
            os.path.abspath(path) for path in include_path_list] + [
                INCLUDE_DIR] 

        # 用来记录，当前的运行模式是否是交互式的
        self.interactive_cli = False

        # 用于记录所有被宏指令 #define 出来的变量
        self.var_set = VarSet()

    # 返回值表示是否是第一次加载
    # 不是第一次加载时候跳过
    # 这里的 filepath 必须使用绝对路径
    def include_file(self, filepath:str) -> bool:
        if not os.path.isabs(filepath): # 应当使用绝对路径
            raise ValueError()
        if not os.path.isfile(filepath): # 文件不存在，抛出异常
            raise FileNotFoundError()
        if filepath in self.exists_path: # 不是第一次加载
            return False
        # 是第一次加载
        self.exists_path.add(filepath)
        new_cmd_list:list[CommandWrap] = []
        all_lines = list(open(filepath, "r", encoding="utf-8"))
        line_id_raw = 0
        while line_id_raw < len(all_lines):
            line = all_lines[line_id_raw].strip()
            cnt = 1 # 记录目前的内容包括了多少行

            # 如果一行内容的末尾是右斜杠
            # 那么这一行应该和下一行连接起来视为一体的
            while line != "" and line[-1] == "\\" and line_id_raw + cnt < len(all_lines):
                line = line[:-1] + all_lines[line_id_raw + cnt].strip()
                cnt += 1
            while line != "" and line[-1] == "\\":
                line = line[:-1]
            if line.find("//") != -1:                          # 字符串中能够找到注释信息
                line = line.split("//", maxsplit=1)[0].strip() # 删除注释信息

            # 行号，要求从 1 开始
            # 多行内容合并后以第一行内容为准
            line_id = line_id_raw + 1 
            new_cmd_list.append(CommandWrap(filepath, line_id, line))

            # 跳过中间合并的行
            line_id_raw += cnt
        self.cmd_list = new_cmd_list + self.cmd_list # 将新命令从头部追加到命令序列
        return True
    
    # 当前命令有可能来自命令行输入而不是文件
    def get_dirnow(self, filepath_str:str) -> str:
        if filepath_str == "<STDIN>": # 获取当前工作目录作为 DIRNOW
            return os.path.abspath(os.getcwd())
        else:
            return os.path.dirname(os.path.abspath(filepath_str))

    # 在所有可以使用的 include_path 中找到第一个匹配项目
    # cmd_file_now 是当前正在执行的命令所在的文件
    def get_first_match_dir(self, filename:str, cmd_file_now:str) -> Optional[str]:
        for path_now in self.include_path:
            # 这里的 "." 是一个特殊目录，表示当前脚本所在目录
            # 如果当前脚本是交互式输入的，则该目录是工作目录
            if path_now == ".":
                path_now = self.get_dirnow(cmd_file_now)
            hypo_filepath = os.path.abspath(os.path.join(path_now, filename))
            if os.path.isfile(hypo_filepath):
                return path_now
        return None

    # 假设 self.cmd_list 中刚刚删除的命令是一个 #if 命令
    # 我们在这个序列中找到 #if 失败后的跳转位置
    # 如果找不到则返回 None
    # 这个命令也可以用来找 #else 匹配的 #endif
    # else_cnt_max = 1 表示我们正在处理 if 的跳转
    # else_cnt_max = 0 表示我们正在处理 else 的跳转
    def get_match_else_or_end_if_pos(self, else_cnt_max:int=1) -> Optional[int]:
        if else_cnt_max not in [0, 1]:
            raise AssertionError()
        layer = 1

        # 存储与当前 if 可能匹配的
        else_list = []
        end_if_list = []

        index = 0
        while index < len(self.cmd_list):
            if self.cmd_list[index].command.strip() == "#if":
                layer += 1
            elif self.cmd_list[index].command.strip() == "#endif":
                layer -= 1
            
            # 找到对应的内容
            # else 由于不执行 layer 的减法，因此应当和开始的 if 处于同一层次
            if layer == 1 and self.cmd_list[index].command.strip() == "#else":
                else_list.append(index)

            # endif 由于会对 layer - 1
            # 因此需要找到低一层的 endif 才是匹配的
            if layer == 0 and self.cmd_list[index].command.strip() == "#endif":
                end_if_list.append(index)

            # 没有找到与当前行匹配的 #endif
            index += 1
            if len(end_if_list) > 0: # 足够使用了
                break
        
        # 检查是否找到了恰好足够数目的 endif 和 else
        if len(end_if_list) == 0:
            raise ValueError("No matching \"#endif\" found.")
        if len(else_list) > else_cnt_max:
            raise ValueError("Redundant \"#else\" found.")
        
        # 找到匹配的 else 的行号
        # 如果 if 失败了可以跳到这里去
        if else_cnt_max == 1:
            return else_list[0] if len(else_list) == 1 else end_if_list[0]

        # else 只能跳转到 endif 不能跳转到其他 else
        else:
            return end_if_list[0]
    
    # 检查一个变量名是否是合法的
    # 如果不合法抛出异常
    def check_macro_var_name(self, var_name:str):
        if not bool(re.fullmatch(r'[_A-Za-z\.][_A-Za-z0-9\.]*', var_name)):
            raise ValueError(f"var name \"{var_name}\" is not allowed.")

    # 执行预处理器命令（井号开头的命令）
    # 注意：预处理命令中不会执行任何宏替换
    # 只有非预处理命令才会执行宏替换
    def execute_preprocessor_cmd(self, cmd_now:CommandWrap):
        if self.verbose:
            print("PCMD:", cmd_now.command.strip()) # 特殊命令就输出一个原始命令就行
        
        # 分离命令头部和命令内容
        try:
            first_part, other_part = cmd_now.command.split(maxsplit=1)
            first_part = first_part.strip()
        except:
            first_part = cmd_now.command.strip()
            other_part = ""
        
        # 预处理命令必须是 # 开头的
        if not first_part.startswith("#"):
            raise AssertionError()
        first_part = first_part[1:]

        # 在这里给出所有特殊命令对应的列表
        if first_part == "include":         # 引入新文件命令
            rel_filepath = other_part.strip() # 获取文件相对路径
            if rel_filepath == "":            # 没有指定文件名
                raise ValueError()
            dirnow = self.get_first_match_dir(rel_filepath, cmd_now.filepath)
            if dirnow is None:
                raise FileNotFoundError(f"Can not include file \"{rel_filepath}\".")
            new_filepath = os.path.abspath(os.path.join(dirnow, rel_filepath))
            self.include_file(new_filepath.strip())

        elif first_part == "exit":   # 直接停止程序
            self.cmd_list.clear()    # 通过清空后续代码的方式停止程序 (文件模式)

            # 如果是交互式命令行模式，则输出再见
            if self.interactive_cli: 
                # 注意，交互式命令行只有在读取文件中的 #exit 时才会遇到这种情况
                print(f"Bye. \n(#exit from {cmd_now.filepath}:{cmd_now.line_id})")
            self.interactive_cli = False 

        # 预处理判断语句
        # 用法：#if value_1 value_2
        # 如果 value_2 的计算结果，可以和 value_1 中给出的模式匹配
        # 因此 value_1 和 value_2 必须都是括号表达式
        # 则继续执行
        # 否则跳转到对应的 #else 或者 #endif
        elif first_part == "if": 
            value_term = ValueTerm.deserialize(f"[{other_part}]")
            if not isinstance(value_term.value, Sequence) or len(value_term.value.objects) != 2:
                raise TypeError("\"#if\" requires exactly two parenthesized expressions.")
            
            # 提取出模板和值
            tmplate_term, brace_seq = value_term.value.objects
            if not isinstance(tmplate_term, BraceSequence) or not isinstance(brace_seq, BraceSequence):
                raise TypeError("\"#if\" requires exactly two parenthesized expressions.")
            
            # 检查是否能够完整匹配
            dic = dict()
            value_term = ValueTerm()     # 构建一个恰好包含一个 BraceSequence 的 ValueTerm
            value_term.value = brace_seq # type:ignore
            value_term = self.rule_set.calc(value_term, self.verbose, self.step_mode)

            # 看前面给出的模板是否能和后面的内容匹配
            if isinstance(value_term.value, BraceSequence):
                flag = match_brace_sequence(tmplate_term, value_term.value, dic)
            else:
                flag = False

            # 由于条件正确，什么都不用做
            # 让程序继续执行即可
            if flag:
                if self.verbose:
                    print("INFO:", "\"#if\" match failed.") 

            # 如果条件不正确
            # 那么需要跳转到下一个 #else 或者 #endif 后
            else:
                # else_cnt_max=1 的含义是，最多可以吸收一个 else （当前语句是 if）
                skip_index = self.get_match_else_or_end_if_pos(else_cnt_max=1)
                if skip_index is None:
                    raise ValueError("No matching \"#else\" or \"#endif\" found.")
                # 中间的命令暴力跳过不执行
                self.cmd_list = self.cmd_list[skip_index+1:] 

                if self.verbose:
                    print("INFO:", f"{skip_index+1} lines skipped by \"#if\".")
        
        # 执行 #else 时，直接跳转到对应的 #endif 处即可
        elif first_part == "else":
            # else_cnt_max=10 的含义是，不能吸收else （当前语句是 else）
            skip_index = self.get_match_else_or_end_if_pos(else_cnt_max=0)
            if skip_index is None:
                raise ValueError("No matching \"#endif\" found.")
            # 中间的命令暴力跳过不执行
            self.cmd_list = self.cmd_list[skip_index+1:] 

            if self.verbose:
                print("INFO:", f"{skip_index+1} lines skipped by \"#else\".")

        # 该语句没有任何功能
        elif first_part == "endif":
            pass
        
        # 从 macro 栈中拿出一个元素
        # 并将其存入变量
        elif first_part == "pop":
            var_name = other_part.strip()
            self.check_macro_var_name(var_name)
            self.var_set.pop_var(var_name)

        # 把某个指定变量的值 push 到 macro 栈中
        # 如果这个变量中还没有值，会报错
        elif first_part == "push":
            var_name = other_part.strip()
            self.check_macro_var_name(var_name)
            self.var_set.push_var(var_name)

        # 把某个表达式的值送给一个变量
        # save 的功能是，先对某个表达式进行计算，然后再把值放进变量
        elif first_part == "save":
            suc = None
            try:
                var_name, exp_val = other_part.split(maxsplit=1)
                suc = True
            except:
                var_name, exp_val = "", ""
                suc = False
            if not suc:
                raise ValueError("there should be two parameters after \"#save\".")
            var_name = var_name.strip()
            self.check_macro_var_name(var_name)
            # 保证表达式的值不是空字符
            # 想要空字符串您用 define 啊，别用 save
            exp_val = exp_val.strip()
            if exp_val == "":
                raise ValueError("There should be two parameters after \"#save\".")
            tmp_value_term = ValueTerm.deserialize(f"[{exp_val}]")
            tmp_value_term = self.rule_set.calc(tmp_value_term, self.verbose, self.step_mode)
            # 记得去掉中括号
            self.var_set.define_var(var_name, tmp_value_term.serialize()[1:-1])

        # erase 的功能是用来在规则集合中，删除某个指定前缀的所有规则
        elif first_part == "erase":
            prefix = other_part.strip()
            if prefix == "":
                raise ValueError("There should be one paramter after \"#erase\".")
            if prefix == "()": # prefix = () 表示删除所有特殊命令（无合法前缀命令）
                prefix = ""    # 在数据库中，无合法前缀命令用 "" 作为前缀
            self.rule_set.erase_rule(prefix)
        
        # 用于定义预处理替换变量
        elif first_part == "define":
            # 从命令中获取变量名称与其中存储的信息
            try:
                var_name, var_value = other_part.split(maxsplit=1)
            except:
                var_name = other_part
                var_value = ""
            var_name = var_name.strip()
            var_value = var_value.strip()
            if var_name == "": # 变量名为空
                raise ValueError("\"#define\" should followed by a var name.")
            self.check_macro_var_name(var_name)
            self.var_set.define_var(var_name, var_value)

        # 删除 define 定义出来的变量
        elif first_part == "undef":
            var_name = other_part.strip()
            self.check_macro_var_name(var_name)
            self.var_set.undef_var(var_name)

        # 显示所有匹配规则
        elif first_part == "rule":
            if other_part.strip() == "":
                print(self.rule_set.show_rules(None))

            else:
                if other_part.strip() == "()": # 通过 #rule () 命令显示所有杂项
                    print(self.rule_set.show_rules(""))
                else:
                    if not SimpleTerm.is_const_val(other_part.strip()):
                        raise ValueError("parameter after \"#rule\" should be const value or ().")
                    print(self.rule_set.show_rules(other_part.strip()))

        # 没有匹配到相关命令
        else:
            raise ValueError(f"Preprocessor command \"#{first_part}\" not found!")

    # 执行指定的命令
    # 注意：只有非预处理语句才需要执行宏替换
    def execute_cmd(self, first_cmd:CommandWrap):

        # 开始考虑当前命令的执行
        # 注：井号开头的行不会做预处理替换
        if first_cmd.command.strip() == "": # 跳过空行
            return
        
        elif first_cmd.command.startswith("//"): # 跳过注释
            return

        # 对命令本身进行宏替换
        # 宏替换过程在所有东西之前做
        # 注意：只有 #define 命令特殊
        #   define 所在的行不会进行 \n 到换行符的替换
        #   其他命令所在的行都需要进行 \n 到换行符的替换
        if not first_cmd.command.strip().startswith("#def"):

            # flag 表示是否成功进行了一次预处理命令替换
            flag, preprocessor_output = self.var_set.solve_once(first_cmd.command)

            # 如果发现确实至少进行了一次替换
            # 则不予执行，重新把命令放入队首
            if flag:
                arr = []
                preprocessor_output_lines = preprocessor_output.split("\n")
                for line in preprocessor_output_lines:
                    arr.append(CommandWrap(first_cmd.filepath, first_cmd.line_id, line))
                if self.verbose:
                    print("\nPREI:", first_cmd.command)
                    for idx, line in enumerate(preprocessor_output_lines): # 一般来说只有两行
                        print(f"PREO[{idx+1}]:", line)
                    if self.step_mode:
                        input("(continue?) ") # 等待用户使用回车继续执行
                self.cmd_list = arr + self.cmd_list
                return # 这里必须 return 否则将会错误地执行一条命令
            
            # 如果预处理器没有进行任何替换
            # 则说明可以正常执行，继续执行即可

        # 执行预处理命令
        if first_cmd.command.startswith("#"): 
            self.execute_preprocessor_cmd(first_cmd)

        # 执行非预处理命令
        # 比如新建规则，或者输出计算结果
        else:
            # 新增匹配规则命令
            if first_cmd.command.find("=>") != -1: 
                if self.verbose:
                    print("NEWR:", first_cmd.command)
                self.rule_set.add_rule( # 将规则加入规则集合
                    ValueMap.deserialize(first_cmd.command), 
                    first_cmd.filepath, 
                    first_cmd.line_id)

            # 执行命令并输出计算结果
            else: 
                value_term = ValueTerm.deserialize(f"[{first_cmd.command.strip()}]")
                if value_term.get_one_var() is not None:
                    raise ValueError(f"No variables (like \"{value_term.get_one_var()}\") allowed in the current expression.")
                new_value_term = self.rule_set.calc(value_term, self.verbose, self.step_mode)

                # 输出前，需要去掉 ValueTerm 的中括号
                print(new_value_term.serialize().strip()[1:-1]) 

    # 执行一条命令
    # 成功执行返回 True, 没有可执行命令返回 False
    # 执行命令时出错抛出异常
    def execute_one(self) -> bool:
        if len(self.cmd_list) == 0:
            return False
        # 获取第一条命令
        first_cmd, self.cmd_list = self.cmd_list[0], self.cmd_list[1:]
        if not isinstance(first_cmd, CommandWrap):
            raise TypeError()
        try:
            self.execute_cmd(first_cmd)
        except Exception as e:
            if self.extra_error_info or str(e) == "" or self.verbose: # 报错信息太简单时候必须提供错误理由
                extra_info = f"\n{traceback.format_exc()}"
            else:
                extra_info = ""
            raise Exception(f"{first_cmd.filepath}:{first_cmd.line_id} {type(e).__name__}: {str(e)} {extra_info}")
        return True
    
    # 执行所有命令
    # 知道出现错误或者执行结束
    def execute_all(self):
        while True:
            flag = self.execute_one()
            if not flag: # 说明已经没有命令可以执行了
                break
    
    # 启动一个交互式用户界面
    def interactive_ui(self):
        self.interactive_cli = True # 表示进入交互式命令模式
        line_id = 1

        # 文件中的 exit 命令通过 self.interactive_cli = False 退出交互式命令行
        while self.interactive_cli: 

            try:
                cmd = input(">>> ").strip()
                if cmd == "#exit": # 退出执行
                    print(f"Bye. \n(#exit from <STDIN>:{line_id})")
                    break

                # 由于 #if 需要向后查询后文的代码
                # 但是我们的交互式命令输入暂时不支持多行输入
                # 因此我们暂时禁止命令行使用这个命令（将来可能可以解禁）
                if len(cmd.split()) >= 1 and cmd.split()[0] == "#if":
                    print(f"<STDIN>:{line_id} ValueError: \"#if\" is not allowed in interactive mode.")
                    continue

                self.cmd_list.append(CommandWrap("<STDIN>", line_id, cmd.strip()))

                # 空行和注释不增加编号
                if cmd.strip() == "" or cmd.strip().startswith("//"):
                    continue

                try:
                    # 执行所有出现的命令，直到无法执行为止
                    self.execute_all()
                except Exception as e:    # 遇到报错，输出错误信息
                    print(e)              # 当出现报错时，清空命令区域
                    self.cmd_list.clear() # 后续可以继续交互
                
                # 只有在正确执行时编号才增加
                line_id += 1
            except KeyboardInterrupt:
                print("\nKeyboardInterrupt")
                print("Use #exit to quit xulang interactive CLI.")
            except Exception as e:
                if str(e) == "": # 如果没有错误信息
                    traceback.print_exc()
                else: # 有错误信息则输出错误信息
                    print(f"<STDIN>:{line_id} {type(e).__name__}: {str(e)}")
        
        # 退出交互模式
        self.interactive_cli = False

    # 注意：需要使用绝对路径调用 filepath
    def run_file(self, filepath:str):
        try:
            self.include_file(filepath)
            try:
                # 执行所有出现的命令，直到无法执行为止
                self.execute_all()
            except Exception as e:    # 遇到报错，输出错误信息
                print(e)              # 当出现报错时，清空命令区域
                self.cmd_list.clear() # 退出程序
            print("(Execution finished normally.)") # 显示出执行正常结束

        except KeyboardInterrupt:
            print("(Execution killed by user.)")

if __name__ == "__main__":
    file_runner = FileRunner([])
    file_runner.interactive_ui()



import re
from typing import Optional

# 描述宏替换过程中使用到的变量集合
# 记录了每个变量的取值，以及一个栈空间（栈中只有值，没有变量名）
class VarSet:
    def __init__(self) -> None:

        # 记录所有宏变量的值
        # 注意：宏变量的值中可能含有换行符
        self.var_map:dict[str, str] = dict()

        # 维护一个变量值构成的栈
        # 可以在变量名冲突的前提下缓存变量信息
        self.var_val_stack:list[str] = []

    # 定义一个变量
    def define_var(self, var_name:str, var_value:str):
        self.var_map[var_name] = var_value

    # 删除一个变量
    def undef_var(self, var_name:str):
        if self.var_map.get(var_name) is not None:
            del self.var_map[var_name]

    # 把一个指定变量的值送到栈里
    def push_var(self, var_name:str):
        if self.var_map.get(var_name) is None:
            raise ValueError(f"Macro var {var_name} is not defined.")
        
        # 列表尾部是栈顶
        self.var_val_stack.append(self.var_map[var_name])

    # 从堆栈里取出一个值，存入指定的变量名
    def pop_var(self, var_name:str):
        if len(self.var_val_stack) == 0:
            raise ValueError("Macro stack is empty, can not pop.")
        self.var_map[var_name] = self.var_val_stack.pop()

    # 找到初次匹配的位置
    # 如果找不到返回 None
    @classmethod
    def match_first_or_empty(cls, pattern: str, text: str) -> Optional[str]:
        match = re.search(pattern, text)
        return match.group() if match else None

    # 找到可能可以替换的变量
    # 如果找到了则连着 ${...} 一起返回，如果找不到则返回 None
    @classmethod
    def what_can_replace(cls, s:str) -> Optional[str]:
        return cls.match_first_or_empty(r"\${[^{}\s]+}", s)

    # 只进行一次必要的替换，其他操作都不做
    # 如果含有字符型换行符，则替换成真换行符
    # 返回值的 bool 代表是否成功进行了替换操作
    def solve_once(self, s:str) -> tuple[bool, str]:
        if s.find(r"\n") != -1:
            s = s.replace(r"\n", "\n", count=1)
            return True, s
        if (can_replace := VarSet.what_can_replace(s)) is not None:
            var_name = can_replace[2:-1] # 删除前面的 "${" 和后面的 "}"
            if self.var_map.get(var_name) is None:
                raise ValueError(f"Macro var \"${{{var_name}}}\" is not defined.")
            s = s.replace(can_replace, self.var_map[var_name])
            return True, s
        return False, s

if __name__ == "__main__":
    var_set = VarSet()
    var_set.define_var("x_1", "val_1")
    var_set.define_var("x_2", "val_2")
    var_set.define_var("d", "2")
    print(var_set.solve_once(r"${x_${d}}"))



import string
from typing import Optional

class SimpleTerm:
    def __init__(self) -> None:
        self.content = "simple_term" # 具体内容名称
        self.has_star = False        # 前缀 star

    @classmethod
    def init(cls, msg:str, has_star:bool) -> 'SimpleTerm':
        if not isinstance(has_star, bool):
            raise TypeError()
        if not isinstance(msg, str) or len(msg) == 0:
            raise TypeError()
        if msg == "_":
            raise ValueError('"_" is not an available identifier.') # 不能只留一个下划线
        if msg == ".":
            raise ValueError('"." is not an available identifier.') # 不能只留一个下划线
        if msg.find("..") != -1:
            raise ValueError(f'No consecutive dots in identifier. (see \"{msg}\")')
        if msg.startswith("."):
            raise ValueError(f'No leading dot in identifier. (see \"{msg}\")')
        if msg.endswith("."):
            raise ValueError(f"No trailing dot in identifier. (see \"{msg}\")")
        for i in range(len(msg)):
            if msg[i] not in (["_", "."] + [
                    char_now for char_now in string.ascii_letters
                ] + [
                    str(num_now) for num_now in range(10)
                ]):
                raise ValueError(f"Illegal character '{msg[i]}'.") # 字母数字下划线
            
        new_item = SimpleTerm()
        new_item.content = msg
        new_item.has_star = has_star
        return new_item
    
    def serialize(self) -> str:
        head = "*" if self.has_star else ""
        return head + self.content

    def json_obj(self) -> dict:
        return {
            "type": "SimpleTerm",
            "content": self.content,
            "has_star": self.has_star
        }

    @classmethod
    def deserialize(cls, s:str) -> 'SimpleTerm':
        s = s.strip()
        has_star = s.startswith("*")
        if has_star:
            s = s[1:] # 分离开头的 star
        return SimpleTerm.init(s, has_star)

    @classmethod
    def from_json_obj(cls, json_obj:dict) -> 'SimpleTerm':
        if json_obj.get("type") != "SimpleTerm":
            raise TypeError()
        if not isinstance(json_obj.get("content"), str):
            raise TypeError()
        if not isinstance(json_obj.get("has_star"), bool):
            raise TypeError()
        return SimpleTerm.init(
            json_obj["content"], json_obj["has_star"])
    
    # 大写字母或者数字开头的符号串
    # 视为常量
    @classmethod
    def is_const_val(cls, s:str) -> bool:
        if s == "":
            raise AssertionError()
        return s[0] in string.ascii_uppercase or s[0] in string.digits
    
    # 如果当前变量不是常量，则返回自己的序列化名
    # 如果当前变量是常量，则返回 None
    def get_one_var(self) -> Optional[str]:
        if SimpleTerm.is_const_val(self.serialize()):
            return None
        else:
            return self.serialize()
    
    # 一种简单的使用 tuple 和 list of str 表示数据的方式
    # 在 python 传参接口中使用
    def simple_express(self) -> str:
        return self.serialize()

if __name__ == "__main__":
    term = SimpleTerm.init("hello", False)
    print(term.serialize())
    print(term.json_obj())
    print(SimpleTerm.deserialize(term.serialize()).json_obj())
